"""This module is responsible for the establishment and configuration of a
Celery instance."""

import logging

from celery import Celery, current_app
from celery.signals import after_task_publish
from flask import Flask

from fractalis.utils import list_classes_with_base_class


logger = logging.getLogger(__name__)


# https://stackoverflow.com/questions/9824172/find-out-whether-celery-task-exists
@after_task_publish.connect
def update_submitted_state(sender, headers, **kwargs):
    """Add 'SUBMITTED' state to celery task."""
    # the task may not exist if sent using `send_task` which
    # sends tasks by name, so fall back to the default result backend
    # if that is the case.
    task = current_app.tasks.get(sender)
    backend = task.backend if task else current_app.backend
    backend.store_result(task_id=headers['id'],
                         result=None,
                         state='SUBMITTED')


def make_celery(app: Flask) -> Celery:
    """Create a celery instance which executes its tasks in the application
    context of our service.
    :param app: The instance of our web service. This holds our configuration.
    :return A celery instance that can submit tasks.
    """
    celery = Celery(app.import_name,
                    backend=app.config['CELERY_RESULT_BACKEND'],
                    broker=app.config['BROKER_URL'])
    celery.conf.update(app.config)

    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with app.app_context():
                return TaskBase.__call__(self, *args, **kwargs)
    celery.Task = ContextTask

    return celery


def register_tasks() -> None:
    """Register all of our Task classes with celery."""
    from fractalis import celery
    from fractalis.analytics.task import AnalyticTask
    from fractalis.data.etl import ETL

    logger.info("Registering ETLs ...")
    etl_classes = list_classes_with_base_class('fractalis.data.etls', ETL)
    for etl_class in etl_classes:
        logger.info("Registering task: {}".format(etl_class.name))
        celery.tasks.register(etl_class)

    logger.info("Registering analysis tasks ...")
    analytics_task_classes = list_classes_with_base_class(
        'fractalis.analytics.task', AnalyticTask)
    for analytics_task_class in analytics_task_classes:
        logger.info("Registering task: {}".format(analytics_task_class.name))
        celery.tasks.register(analytics_task_class)



"""This module provides tests for the janitor"""

import os
import json
from pathlib import Path
from shutil import rmtree

from fractalis.cleanup import janitor
from fractalis import app, redis, celery


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestCleanup:

    data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')

    def teardown_method(self, method):
        redis.flushall()
        rmtree(self.data_dir)

    def test_janitor_removes_untracked_files(self):
        os.makedirs(self.data_dir, exist_ok=True)
        Path(os.path.join(self.data_dir, 'abc')).touch()
        janitor()
        assert not os.path.exists(os.path.join(self.data_dir, 'abc'))

    def test_janitor_does_not_remove_tracked_files(self):
        os.makedirs(self.data_dir, exist_ok=True)
        Path(os.path.join(self.data_dir, 'abc')).touch()
        redis.set('data:abc', '')
        janitor()
        assert os.path.exists(os.path.join(self.data_dir, 'abc'))

    def test_janitor_removes_redis_keys_without_file(self, monkeypatch):
        os.makedirs(self.data_dir, exist_ok=True)
        path = os.path.join(self.data_dir, '123')
        data_state = {
            'file_path': path
        }
        redis.set('data:123', json.dumps(data_state))

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUCCESS'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        janitor()
        assert not redis.exists('data:123')

    def test_janitor_doesnt_remove_redis_keys_with_file(self, monkeypatch):
        os.makedirs(self.data_dir, exist_ok=True)
        path = os.path.join(self.data_dir, '123')
        Path(path).touch()
        data_state = {
            'file_path': path
        }
        redis.set('data:123', json.dumps(data_state))

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUCCESS'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        janitor()
        assert redis.exists('data:123')

    def test_janitor_doesnt_remove_redis_keys_if_not_finished(
            self, monkeypatch):
        os.makedirs(self.data_dir, exist_ok=True)
        path = os.path.join(self.data_dir, '123')
        data_state = {
            'file_path': path
        }
        redis.set('data:123', json.dumps(data_state))

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUBMITTED'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        janitor()
        assert redis.exists('data:123')



"""This module provides TransmartHandler, an implementation of ETLHandler for
tranSMART."""

import logging

import requests

from fractalis.data.etlhandler import ETLHandler


logger = logging.getLogger(__name__)


class TransmartHandler(ETLHandler):
    """This ETLHandler provides integration with tranSMART.

    'tranSMART is a knowledge management platform that enables scientists to
    develop and refine research hypotheses by investigating correlations
    between genetic and phenotypic data, and assessing their analytical results
    in the context of published literature and other work.'

    Project URL: https://github.com/transmart
    """

    _handler = 'transmart'

    @staticmethod
    def make_label(descriptor: dict) -> str:
        return descriptor['label']

    def _get_token_for_credentials(self, server: str, auth: dict) -> str:
        try:
            user = auth['user']
            passwd = auth['passwd']
            if len(user) == 0 or len(passwd) == 0:
                raise KeyError
        except KeyError:
            error = "The authentication object must contain the non-empty " \
                    "fields 'user' and 'passwd'."
            logger.error(error)
            raise ValueError(error)
        r = requests.post(url=server + '/oauth/token',
                          params={
                              'grant_type': 'password',
                              'client_id': 'glowingbear-js',
                              'client_secret': '',
                              'username': user,
                              'password': passwd
                          },
                          headers={'Accept': 'application/json'},
                          timeout=10)
        if r.status_code != 200:
            error = "Could not authenticate. " \
                    "Reason: [{}]: {}".format(r.status_code, r.text)
            logger.error(error)
            raise ValueError(error)
        try:
            response = r.json()
            return response['access_token']
        except Exception:
            error = "Could not authenticate. " \
                    "Got unexpected response: '{}'".format(r.text)
            logger.error(error)
            raise ValueError(error)



"""This module provides sample data."""

import pandas as pd
import numpy as np
import string
import random

from fractalis.data.etl import ETL


class RandomNumericalArrayETL(ETL):

    name = 'test_numerical_array_etl'
    produces = 'numerical_array'

    @staticmethod
    def can_handle(handler: str, descriptor: dict) -> bool:
        return handler == 'test' and \
               descriptor['data_type'] == 'numerical_array'

    def extract(self, server: str,
                token: str, descriptor: dict) -> pd.DataFrame:
        features = [''.join(random.choice(string.ascii_letters + string.digits)
                            for _ in range(10))
                    for _ in range(descriptor['num_features'])]
        data = pd.DataFrame(np.random.randn(
            descriptor['num_samples'], descriptor['num_features']).tolist(),
                            columns=features)
        return data

    def transform(self, raw_data: pd.DataFrame,
                  descriptor: dict) -> pd.DataFrame:
        raw_data.insert(0, 'id', raw_data.index.astype('str'))
        df = pd.melt(raw_data, id_vars='id', var_name='feature')
        return df



"""This module provides functions and celery tasks used to cleanup expired
items. It is also used to keep several components synchronized, e.g. the redis
db and the file system.
"""

import os
import json
import logging
from shutil import rmtree

from fractalis import redis, app, celery


logger = logging.getLogger(__name__)


def remove_data(task_id: str) -> None:
    """Remove all traces of any data associated with the given id. That includes
    redis and the file system.
    :param task_id: The id associated with a data state
    :param wait: Wait for all subtasks to finish before returning
    """
    key = 'data:{}'.format(task_id)
    value = redis.get(key)
    celery.control.revoke(task_id, terminate=True, signal='SIGUSR1')
    redis.delete(key)
    if value:
        data_state = json.loads(value)
        remove_file(data_state['file_path'])
    else:
        logger.warning("Can't delete file for task id '{}',because there is "
                       "no associated entry in Redis.".format(task_id))


def remove_file(file_path: str) -> None:
    """Remove the file for the given file path.
    :param file_path: Path of file to remove.
    """
    try:
        os.remove(file_path)
    except FileNotFoundError:
        logger.warning("Attempted to remove file '{}', "
                       "but it does not exist.".format(file_path))


def cleanup_all() -> None:
    """Reset redis, celery and the filesystem. This is only useful for testing
    and should !!!NEVER!!! be used for anything else.
    """
    celery.control.purge()
    for key in redis.keys('data:*'):
        value = redis.get(key)
        try:
            data_state = json.loads(value)
        except ValueError:
            continue
        task_id = data_state.get('task_id')
        if task_id is not None:
            async_result = celery.AsyncResult(task_id)
            if async_result.state == 'SUBMITTED':
                async_result.get(propagate=False)
    redis.flushall()
    tmp_dir = app.config['FRACTALIS_TMP_DIR']
    if os.path.exists(tmp_dir):
        rmtree(tmp_dir)
    assert not os.path.exists(tmp_dir)



"""This module contains tests for the survival module."""

from lifelines.datasets import load_waltons

from fractalis.analytics.tasks.survival.main import SurvivalTask


class TestSurvivalTask:

    task = SurvivalTask()

    def test_correct_output_for_simple_input(self):
        df = load_waltons()
        df.insert(0, 'id', df.index)
        duration = df[['id', 'T']].copy()
        duration.insert(1, 'feature', 'duration')
        duration.columns.values[2] = 'value'
        results = self.task.main(durations=[duration],
                                 categories=[],
                                 event_observed=[],
                                 estimator='KaplanMeier',
                                 id_filter=[],
                                 subsets=[])
        assert results['label'] == 'duration'
        assert len(results['categories']) == 1
        assert len(results['subsets']) == 1
        assert results['stats'][''][0]['timeline']
        assert results['stats'][''][0]['estimate']
        assert results['stats'][''][0]['ci_lower']
        assert results['stats'][''][0]['ci_upper']

    def test_correct_output_for_complex_input(self):
        df = load_waltons()
        df.insert(0, 'id', df.index)
        duration = df[['id', 'T']].copy()
        duration.insert(1, 'feature', 'duration')
        duration.columns.values[2] = 'value'
        event_observed = df[['id', 'E']].copy()
        event_observed.insert(1, 'feature', 'was_observed')
        categories = df[['id', 'group']].copy()
        categories.insert(1, 'feature', 'group')
        results = self.task.main(durations=[duration],
                                 categories=[categories],
                                 event_observed=[event_observed],
                                 estimator='NelsonAalen',
                                 id_filter=[],
                                 subsets=[])
        assert results['label'] == 'duration'
        assert len(results['categories']) == 2
        assert len(results['subsets']) == 1
        assert results['stats']['control'][0]['timeline']
        assert results['stats']['control'][0]['estimate']
        assert results['stats']['control'][0]['ci_lower']
        assert results['stats']['control'][0]['ci_upper']
        assert results['stats']['miR-137'][0]['timeline']
        assert results['stats']['miR-137'][0]['estimate']
        assert results['stats']['miR-137'][0]['ci_lower']
        assert results['stats']['miR-137'][0]['ci_upper']

    def test_can_handle_nans(self):
        df = load_waltons()
        df.insert(0, 'id', df.index)
        duration = df[['id', 'T']].copy()
        duration.insert(1, 'feature', 'duration')
        duration.columns.values[2] = 'value'
        duration.loc[duration.index % 2 == 0, 'value'] = float('nan')
        self.task.main(durations=[duration],
                       categories=[],
                       event_observed=[],
                       estimator='KaplanMeier',
                       id_filter=[],
                       subsets=[])

    def test_can_handle_empty_groups(self):
        df = load_waltons()
        df.insert(0, 'id', df.index)
        subset1 = df[df['group'] == 'control']['id'].tolist()
        subset2 = df[df['group'] == 'miR-137']['id'].tolist()
        duration = df[['id', 'T']].copy()
        categories = df[['id', 'group']].copy()
        duration.insert(1, 'feature', 'duration')
        categories.insert(1, 'feature', 'group')
        duration.columns.values[2] = 'value'
        categories.columns.values[2] = 'value'
        results = self.task.main(durations=[duration],
                                 categories=[categories],
                                 event_observed=[],
                                 estimator='KaplanMeier',
                                 id_filter=[],
                                 subsets=[subset1, subset2])
        assert not results['stats']['miR-137'].get(0)
        assert not results['stats']['control'].get(1)



"""This module tests the analytics controller module."""

import json
import time
import uuid
from uuid import uuid4

import flask
import pytest

from fractalis import sync, app


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestAnalytics:

    @pytest.fixture(scope='function')
    def test_client(self):
        sync.cleanup_all()
        from fractalis import app
        app.testing = True
        with app.test_client() as test_client:
            yield test_client
            sync.cleanup_all()

    @pytest.fixture(scope='function')
    def small_data_post(self, test_client):
        return lambda random, wait=0: test_client.post(
            '/data?wait={}'.format(wait), data=flask.json.dumps(dict(
                handler='test',
                server='localhost:1234',
                auth={'token': '7746391376142672192764'},
                descriptors=[
                    {
                        'data_type': 'default',
                        'concept': str(uuid4()) if random else 'concept'
                    }
                ]
            )))

    def test_new_resource_created(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='add_test_task',
            args={'a': 1, 'b': 2}
        )))
        body = flask.json.loads(rv.get_data())
        assert rv.status_code == 201, body
        new_url = '/analytics/{}'.format(body['task_id'])
        assert uuid.UUID(body['task_id'])
        assert test_client.head(new_url).status_code == 200

    @pytest.fixture(scope='function',
                    params=[{'task_name': 'i_dont_exist_task',
                             'args': {'a': 1, 'b': 2}},
                            {'task_name': '',
                             'args': {'a': 1, 'b': 2}}])
    def bad_request(self, test_client, request):
        return test_client.post('/analytics',
                                data=flask.json.dumps(request.param))

    def test_400_if_post_body_invalid(self, bad_request):
        assert bad_request.status_code == 400

    def test_no_conflict_when_running_task_twice(self, test_client):
        rv1 = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='nothing_test_task',
            args={'seconds': 4}
        )))
        assert rv1.status_code == 201

        rv2 = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='nothing_test_task',
            args={'seconds': 4}
        )))
        assert rv2.status_code == 201

        body1 = flask.json.loads(rv1.get_data())
        new_url1 = '/analytics/{}?wait=0'.format(body1['task_id'])
        new_response1 = test_client.get(new_url1)
        assert new_response1.status_code == 200
        new_body1 = flask.json.loads(new_response1.get_data())
        assert new_body1['state'] != 'FAILURE'

        body2 = flask.json.loads(rv2.get_data())
        new_url2 = '/analytics/{}?wait=0'.format(body2['task_id'])
        new_response2 = test_client.get(new_url2)
        assert new_response2.status_code == 200
        new_body2 = flask.json.loads(new_response2.get_data())
        assert new_body2['state'] != 'FAILURE'

    def test_403_if_no_session_auth(self, test_client, small_data_post):
        small_data_post(random=False)
        with test_client.session_transaction() as sess:
            assert len(sess['data_tasks']) == 1
            task_id = sess['data_tasks'][0]
        with test_client.session_transaction() as sess:
            sess['data_tasks'] = []
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='sum_df_test_task',
            args={'a': '${}$'.format(task_id)}
        )))
        body = flask.json.loads(rv.get_data())
        assert rv.status_code == 201, body
        url = '/analytics/{}?wait=1'.format(body['task_id'])
        rv = test_client.get(url)
        body = flask.json.loads(rv.get_data())
        assert rv.status_code == 200, body
        assert body['state'] == 'FAILURE', body
        assert 'PermissionError' in body['result'], body

    def test_resource_deleted(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='add_test_task',
            args={'a': 1, 'b': 1}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        assert test_client.head(new_url).status_code == 200
        assert test_client.delete(new_url).status_code == 200
        assert test_client.head(new_url).status_code == 403

    def test_403_if_deleting_non_existing_resource(self, test_client):
        rv = test_client.delete('/analytics/{}?wait=1'.format(str(uuid4())))
        assert rv.status_code == 403

    def test_running_resource_deleted(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='nothing_test_task',
            args={'seconds': 4}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        assert test_client.head(new_url).status_code == 200
        assert test_client.delete(new_url).status_code == 200
        assert test_client.head(new_url).status_code == 403

    def test_403_if_deleting_without_auth(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='nothing_test_task',
            args={'seconds': 4}
        )))
        assert rv.status_code == 201
        time.sleep(1)
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        with test_client.session_transaction() as sess:
            sess['analytic_tasks'] = []
        assert test_client.delete(new_url).status_code == 403

    def test_status_contains_result_if_finished(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='add_test_task',
            args={'a': 1, 'b': 2}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        new_response = test_client.get(new_url)
        assert new_response.status_code == 200
        new_body = flask.json.loads(new_response.get_data())
        assert new_body['state'] == 'SUCCESS', new_body
        assert json.loads(new_body['result'])['sum'] == 3, new_body

    def test_status_result_empty_if_not_finished(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='nothing_test_task',
            args={'seconds': 4}
        )))
        time.sleep(1)
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=0'.format(body['task_id'])
        new_response = test_client.get(new_url)
        assert new_response.status_code == 200
        new_body = flask.json.loads(new_response.get_data())
        assert not new_body['result']
        assert new_body['state'] == 'SUBMITTED'

    def test_correct_response_if_task_fails(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='div_test_task',
            args={'a': 2, 'b': 0}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        new_response = test_client.get(new_url)
        assert new_response.status_code == 200
        new_body = flask.json.loads(new_response.get_data())
        assert new_body['state'] == 'FAILURE'
        assert 'ZeroDivisionError' in new_body['result']

    def test_403_if_status_non_existing_resource(self, test_client):
        assert test_client.get('/analytics/{}?wait=1'
                               .format(str(uuid4()))).status_code == 403

    def test_403_if_status_without_auth(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='nothing_test_task',
            args={'seconds': 4}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=0'.format(body['task_id'])
        with test_client.session_transaction() as sess:
            sess['analytic_tasks'] = []
        assert test_client.get(new_url).status_code == 403

    def test_float_when_summing_up_df(self, test_client, small_data_post):
        data_tasks = []

        small_data_post(random=True, wait=1)
        with test_client.session_transaction() as sess:
            data_tasks += sess['data_tasks']
            assert len(data_tasks) == 1

        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='sum_df_test_task',
            args={'a': '${}$'.format(data_tasks[0])}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        new_response = test_client.get(new_url)
        new_body = flask.json.loads(new_response.get_data())
        assert new_response.status_code == 200, new_body
        assert new_body['state'] == 'SUCCESS', new_body
        assert float(json.loads(new_body['result'])['sum'])

    def test_float_when_summing_up_encrypted_df(
            self, test_client, small_data_post):
        app.config['FRACTALIS_ENCRYPT_CACHE'] = True
        data_tasks = []
        small_data_post(random=True, wait=1)
        with test_client.session_transaction() as sess:
            data_tasks += sess['data_tasks']
            assert len(data_tasks) == 1

        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='sum_df_test_task',
            args={'a': '${}$'.format(data_tasks[0])}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        new_response = test_client.get(new_url)
        new_body = flask.json.loads(new_response.get_data())
        assert new_response.status_code == 200, new_body
        assert new_body['state'] == 'SUCCESS', new_body
        assert float(json.loads(new_body['result'])['sum'])
        app.config['FRACTALIS_ENCRYPT_CACHE'] = False

    def test_exception_if_result_not_json(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='invalid_json_task',
            args={'a': 1}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        new_response = test_client.get(new_url)
        assert new_response.status_code == 200
        new_body = flask.json.loads(new_response.get_data())
        assert new_body['state'] == 'FAILURE'
        assert 'TypeError' in new_body['result']

    def test_exception_if_result_not_dict(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='no_dict_task',
            args={'a': 1}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        new_response = test_client.get(new_url)
        assert new_response.status_code == 200
        new_body = flask.json.loads(new_response.get_data())
        assert new_body['state'] == 'FAILURE'
        assert 'TypeError' in new_body['result']

    def test_can_handle_df_list(self, test_client, small_data_post):
        small_data_post(random=True, wait=1)
        small_data_post(random=True, wait=1)
        small_data_post(random=True, wait=1)
        with test_client.session_transaction() as sess:
            data_tasks = sess['data_tasks']
            assert len(data_tasks) == 3
        df_list = ['${}$'.format(data_task_id) for data_task_id in data_tasks]
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='merge_df_task',
            args={'df_list': df_list}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        new_response = test_client.get(new_url)
        assert new_response.status_code == 200
        new_body = flask.json.loads(new_response.get_data())
        assert new_body['state'] == 'SUCCESS', new_body
        assert len(json.loads(json.loads(new_body['result'])['df'])) == 30

    def test_can_handle_empty_df_list(self, test_client):
        rv = test_client.post('/analytics', data=flask.json.dumps(dict(
            task_name='merge_df_task',
            args={'df_list': []}
        )))
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        new_url = '/analytics/{}?wait=1'.format(body['task_id'])
        new_response = test_client.get(new_url)
        assert new_response.status_code == 200
        new_body = flask.json.loads(new_response.get_data())
        assert new_body['state'] == 'SUCCESS', new_body



"""This module provides sample data."""

import pandas as pd
import random

from fractalis.data.etl import ETL


class RandomCategoricalETL(ETL):

    name = 'test_categorical_etl'
    produces = 'categorical'

    @staticmethod
    def can_handle(handler: str, descriptor: dict) -> bool:
        return handler == 'test' and \
               descriptor['data_type'] == 'categorical'

    def extract(self, server: str,
                token: str, descriptor: dict) -> pd.DataFrame:
        data = pd.DataFrame([random.choice(descriptor['values'])
                             for i in range(descriptor['num_samples'])])
        return data

    def transform(self, raw_data: pd.DataFrame,
                  descriptor: dict) -> pd.DataFrame:
        raw_data.insert(0, 'id', raw_data.index.astype('str'))
        df = pd.melt(raw_data, id_vars='id', var_name='feature')
        return df



"""This module provides test for the 'etl' module."""

import json

import pandas as pd

from fractalis import redis
from fractalis.data.etl import ETL


# noinspection PyMissingOrEmptyDocstring
class Request:

    id = 123


# noinspection PyMissingOrEmptyDocstring
class RequestStackDummy:

    top = Request()


# noinspection PyMissingOrEmptyDocstring
class MockETL(ETL):
    @property
    def name(self) -> str:
        return ''

    def transform(self, raw_data: object, descriptor: dict) -> pd.DataFrame:
        pass

    @staticmethod
    def can_handle(handler: str, descriptor: dict) -> bool:
        pass

    def extract(self, server: str, token: str, descriptor: dict) -> object:
        pass

    @property
    def produces(self) -> str:
        return ''


# noinspection PyMissingOrEmptyDocstring, PyMissingTypeHints
class TestETL:

    etl = MockETL()
    etl.request_stack = RequestStackDummy()

    def test_update_redis(self):
        df1 = pd.DataFrame([[1, 2, 3]], columns=['id', 'feature', 'value'])
        df2 = pd.DataFrame([[1, 3]], columns=['id', 'value'])
        df3 = pd.DataFrame([], columns=['id', 'feature', 'value'])
        redis.set('data:123', json.dumps({'meta': {}}))

        self.etl.update_redis(data_frame=df1)
        data_state = json.loads(redis.get('data:123'))
        assert data_state['meta']['features'] == [2]

        self.etl.update_redis(data_frame=df2)
        data_state = json.loads(redis.get('data:123'))
        assert data_state['meta']['features'] == []

        self.etl.update_redis(data_frame=df3)
        data_state = json.loads(redis.get('data:123'))
        assert data_state['meta']['features'] == []



"""This module provides tests for the transmart etl handler."""

import pytest
import responses

from fractalis.data.etls.transmart.handler_transmart import TransmartHandler


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestTransmartHandler:

    @pytest.fixture(scope='function', params=[
        {'server': 'http://foo.bar', 'auth': ''},
        {'server': 'http://foo.bar', 'auth': {}},
        {'server': 'http://foo.bar', 'auth': {'abc': 'abc'}},
        {'server': 'http://foo.bar', 'auth': {'token': ''}},
        {'server': 'http://foo.bar', 'auth': {'token': object}},
        {'server': 'http://foo.bar', 'auth': {'user': 'foo'}},
        {'server': 'http://foo.bar', 'auth': {'passwd': 'foo'}},
        {'server': 'http://foo.bar', 'auth': {'user': '', 'passwd': ''}},
        {'server': 'http://foo.bar', 'auth': {'user': 'foo', 'passwd': ''}},
        {'server': 'http://foo.bar', 'auth': {'user': '', 'passwd': 'foo'}},
        {'server': 'http://foo.bar',
         'auth': {'user': '', 'passwd': '', 'foo': 'bar'}},
        {'server': '', 'auth': {'token': 'foo'}},
        {'server': object, 'auth': {'token': 'foo'}}
    ])
    def bad_init_args(self, request):
        return request.param

    def test_throws_if_bad_init_args(self, bad_init_args):
        with pytest.raises(ValueError):
            TransmartHandler(**bad_init_args)

    def test_returns_token_for_credentials(self):
        with responses.RequestsMock() as response:
            response.add(response.POST, 'http://foo.bar/oauth/token',
                         body='{"access_token":"foo-token","token_type":"bearer","expires_in":43185,"scope":"read write"}',  # noqa: 501
                         status=200,
                         content_type='application/json')
            tmh = TransmartHandler(server='http://foo.bar',
                                   auth={'user': 'foo', 'passwd': 'bar'})
            assert tmh._token == 'foo-token'

    def test_auth_raises_exception_for_non_json_return(self):
        with responses.RequestsMock() as response:
            response.add(response.POST, 'http://foo.bar/oauth/token',
                         body='123{//}',
                         status=200,
                         content_type='application/json')
            with pytest.raises(ValueError) as e:
                TransmartHandler(server='http://foo.bar',
                                 auth={'user': 'foo', 'passwd': 'bar'})
                assert 'unexpected response' in e

    def test_auth_raises_exception_for_non_200_return(self):
        with responses.RequestsMock() as response:
            response.add(response.POST, 'http://foo.bar/oauth/token',
                         body='something',
                         status=400,
                         content_type='application/json')
            with pytest.raises(ValueError) as e:
                TransmartHandler(server='http://foo.bar',
                                 auth={'user': 'foo', 'passwd': 'bar'})
                assert '[400]' in e



from fractalis.utils import list_classes_with_base_class
from .etlhandler import ETLHandler
from .etl import ETL
from .check import IntegrityCheck

HANDLER_REGISTRY = list_classes_with_base_class('fractalis.data.etls',
                                                ETLHandler)
ETL_REGISTRY = list_classes_with_base_class('fractalis.data.etls',
                                            ETL)
CHECK_REGISTRY = list_classes_with_base_class('fractalis.data',
                                              IntegrityCheck)



"""Provides categorical concept ETL for tranSMART."""

from fractalis.data.etls.transmart.shared import create_etl_type, CATEGORICAL_FIELD

CategoricalETL = create_etl_type(
    name_='transmart_categorical_etl',
    produces_='categorical',
    field_name=CATEGORICAL_FIELD
)



"""This module provides tests for the ada etl handler."""

import pytest
import responses

from fractalis.data.etls.ada.handler_ada import AdaHandler


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestAdaHandler:

    @pytest.fixture(scope='function', params=[
        {'server': 'http://foo.bar', 'auth': ''},
        {'server': 'http://foo.bar', 'auth': {}},
        {'server': 'http://foo.bar', 'auth': {'abc': 'abc'}},
        {'server': 'http://foo.bar', 'auth': {'token': ''}},
        {'server': 'http://foo.bar', 'auth': {'token': object}},
        {'server': 'http://foo.bar', 'auth': {'user': 'foo'}},
        {'server': 'http://foo.bar', 'auth': {'passwd': 'foo'}},
        {'server': 'http://foo.bar', 'auth': {'user': '', 'passwd': ''}},
        {'server': 'http://foo.bar', 'auth': {'user': 'foo', 'passwd': ''}},
        {'server': 'http://foo.bar', 'auth': {'user': '', 'passwd': 'foo'}},
        {'server': 'http://foo.bar',
         'auth': {'user': '', 'passwd': '', 'foo': 'bar'}},
        {'server': '', 'auth': {'token': 'foo'}},
        {'server': object, 'auth': {'token': 'foo'}}
    ])
    def bad_init_args(self, request):
        return request.param

    def test_throws_if_bad_init_args(self, bad_init_args):
        with pytest.raises(ValueError):
            AdaHandler(**bad_init_args)

    @staticmethod
    def request_callback(request):
        headers = {
            'Set-Cookie': 'PLAY2AUTH_SESS_ID=foo-token'}
        body = ''
        return 200, headers, body

    def test_returns_token_for_credentials(self):
        with responses.RequestsMock() as response:
            response.add_callback(response.POST, 'http://foo.bar/login',
                                  callback=self.request_callback,
                                  content_type='application/json')
            adah = AdaHandler(server='http://foo.bar',
                              auth={'user': 'foo', 'passwd': 'bar'})
            assert adah._token == 'foo-token'

    def test_auth_raises_exception_for_non_json_return(self):
        with responses.RequestsMock() as response:
            response.add(response.POST, 'http://foo.bar/login',
                         body='123{//}',
                         status=200,
                         content_type='application/json')
            with pytest.raises(ValueError) as e:
                AdaHandler(server='http://foo.bar',
                           auth={'user': 'foo', 'passwd': 'bar'})
                assert 'unexpected response' in e

    def test_auth_raises_exception_for_non_200_return(self):
        with responses.RequestsMock() as response:
            response.add(response.POST, 'http://foo.bar/login',
                         body='something',
                         status=400,
                         content_type='application/json')
            with pytest.raises(ValueError) as e:
                AdaHandler(server='http://foo.bar',
                           auth={'user': 'foo', 'passwd': 'bar'})
                assert '[400]' in e



"""
Documentation: http://stackoverflow.com/questions/24238743/flask-decorator-to-verify-json-and-json-schema  # noqa: 501
"""
from functools import wraps

from flask import request, jsonify
from werkzeug.exceptions import BadRequest
from jsonschema import validate
from jsonschema.exceptions import ValidationError


def validate_json(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            json = request.get_json(force=True)
            if not (json):
                raise BadRequest()
        except BadRequest:
            error_msg = "Request payload must be valid JSON."
            return jsonify({'error': error_msg}), 400
        return f(*args, **kwargs)
    return wrapper


def validate_schema(schema):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            json = request.get_json(force=True)
            try:
                validate(json, schema)
            except ValidationError as e:
                return jsonify({'error': e.message}), 400
            return f(*args, **kwargs)
        return wrapper
    return decorator



"""This module provides AnalyticTask, which is a modification of a standard
Celery task tailored to Fractalis."""
import abc
import json
import re
import logging
from uuid import UUID
from typing import List, Tuple, Union

from pandas import read_pickle, DataFrame
from celery import Task
from Cryptodome.Cipher import AES

from fractalis import redis, app
from fractalis.utils import get_cache_encrypt_key

logger = logging.getLogger(__name__)


class AnalyticTask(Task, metaclass=abc.ABCMeta):
    """AnalyticTask is a tailored Celery Task that enforces a certain pattern
    for all Fractalis analytic tasks and provides certain functionality like
    the parsing of the arguments before submitting it to celery.
    """
    @property
    @abc.abstractmethod
    def name(self) -> str:
        """The name of the task."""
        pass

    @staticmethod
    def factory(task_name: str) -> 'AnalyticTask':
        """Initialize the correct task based on the given arguments.
        :param task_name: The name of the task to initialize.
        :return: An initialized child of this class.
        """
        from . import TASK_REGISTRY
        for task in TASK_REGISTRY:
            if task.name == task_name:
                return task()

    @abc.abstractmethod
    def main(self, *args, **kwargs) -> dict:
        """Since we hijack run(), we need a new entry point for every task.
        This method is called by our modified run method with all parsed
        arguments.
        :return A dict containing the results of the task.
        """
        pass

    @staticmethod
    def secure_load(file_path: str) -> DataFrame:
        """Decrypt data so they can be loaded into a pandas data frame.
        :param file_path: The location of the encrypted file.
        :return: The decrypted file loaded into a pandas data frame.
        """
        key = get_cache_encrypt_key(app.config['SECRET_KEY'])
        with open(file_path, 'rb') as f:
            nonce, tag, ciphertext = [f.read(x) for x in (16, 16, -1)]
        cipher = AES.new(key, AES.MODE_EAX, nonce)
        data = cipher.decrypt_and_verify(ciphertext, tag)
        data = data.decode('utf-8')
        data = json.loads(data)
        df = DataFrame.from_dict(data)
        return df

    def data_task_id_to_data_frame(
            self, data_task_id: str,
            session_data_tasks: List[str], decrypt: bool) -> DataFrame:
        """Attempts to load the data frame associated with the provided data id
        :param data_task_id: The data id associated with the previously loaded
        data.
        :param session_data_tasks: A list of data tasks previously executed by
        this the requesting session. This is used for permission checks.
        :param decrypt: Specify whether the data have to be decrypted for usage
        only part of the data, for instance some genes out of thousands.
        :return: A pandas data frame associated with the data id.
        """
        if data_task_id not in session_data_tasks:
            error = "No permission to use data_task_id '{}' " \
                    "for analysis".format(data_task_id)
            logger.error(error)
            raise PermissionError(error)
        entry = redis.get('data:{}'.format(data_task_id))
        if not entry:
            error = "The key '{}' does not match any entry in Redis. " \
                    "Value probably expired.".format(data_task_id)
            logger.error(error)
            raise LookupError(error)
        data_state = json.loads(entry)
        async_result = self.AsyncResult(data_task_id)
        if async_result.state != 'SUCCESS':
            error = "The data task '{}' has not been loaded, yet. " \
                    "Wait for it to complete before using it in an " \
                    "analysis task.".format(data_task_id)
            logger.error(error)
            raise ValueError(error)
        file_path = data_state['file_path']
        if decrypt:
            return self.secure_load(file_path)
        else:
            df = read_pickle(file_path, compression='gzip')
        return df

    @staticmethod
    def apply_filters(df: DataFrame, filters: dict) -> DataFrame:
        """Apply filter to data frame and return it.
        :param df: The data frame.
        :param filters: The filters where each key is represents a column
        in the data frame and the value a list of values to keep.
        :return: Filtered data frame.
        """
        for key in filters:
            if filters[key]:
                df = df[df[key].isin(filters[key])]
        return df

    @staticmethod
    def contains_data_task_id(value: str) -> bool:
        """Check whether the given string represents
        a special data id arguments.
        :param value: The string to test.
        :return: True if argument contains data_task_id.
        """
        return isinstance(value, str) and \
            value.startswith('$') and \
            value.endswith('$')

    @staticmethod
    def parse_value(value: str) -> Tuple[Union[str, None], dict]:
        """Extract data task id and filters from the string.
        :param value: A string that contains a data task id.
        :return: A tuple of id and filters to apply later.
        """
        value = value[1:-1]
        # noinspection PyBroadException
        try:
            value = json.loads(value)
            data_task_id = str(value['id'])
            filters = value.get('filters')
        except Exception:
            logger.warning("Failed to parse value. "
                           "Fallback assumption is that it contains the id "
                           "but nothing else.")
            data_task_id = value
            filters = None
        # noinspection PyBroadException
        try:
            data_task_id = str(UUID(data_task_id))
        except Exception:
            logger.warning("'{}' is no valid task id.".format(data_task_id))
            data_task_id = None
        return data_task_id, filters

    def prepare_args(self, session_data_tasks: List[str],
                     args: dict, decrypt: bool) -> dict:
        """Replace data task ids in the arguments with their associated
        data frame located on the file system. This currently works for non
        nested strings and non nested lists containing strings.
        :param session_data_tasks: We use this list to check access.
        :param args: The arguments submitted to run().
        :param decrypt: Indicates whether cache must be decrypted to be used.
        :return: The new parsed arguments
        """
        parsed_args = {}
        for arg in args:
            value = args[arg]

            # value is data id
            if self.contains_data_task_id(value):
                data_task_id, filters = self.parse_value(value)
                df = self.data_task_id_to_data_frame(
                    data_task_id, session_data_tasks, decrypt)
                if filters:
                    df = self.apply_filters(df, filters)
                value = df

            # value is list containing data ids
            if (isinstance(value, list) and
                    value and self.contains_data_task_id(value[0])):
                dfs = []
                for el in value:
                    data_task_id, filters = self.parse_value(el)
                    df = self.data_task_id_to_data_frame(
                        data_task_id, session_data_tasks, decrypt)
                    if filters:
                        df = self.apply_filters(df, filters)
                    dfs.append(df)
                value = dfs

            parsed_args[arg] = value

        return parsed_args

    @staticmethod
    def task_result_to_json(result: dict) -> str:
        """Transform task result to JSON so we can send it as a response.
        :param result: The return value of main()
        :return: A string that can be parsed by the front-end for instance.
        """
        try:
            if type(result) != dict:
                error = "The task '{}' returned an object with type '{}', " \
                        "instead of expected type 'dict'."
                logger.error(error)
                raise ValueError(error)
            result = json.dumps(result)
        except TypeError as e:
            logger.exception(e)
            raise
        # NaN is invalid JSON and JS can't parse it. null on the other hand...
        result = re.sub(r'NaN', 'null', result)
        return result

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        """Set lifetime of analysis result to prevent redis from consuming
        too much memory."""
        redis.expire(name='celery-task-meta-{}'.format(task_id),
                     time=app.config['FRACTALIS_RESULT_LIFETIME'])

    def run(self, session_data_tasks: List[str],
            args: dict, decrypt: bool) -> str:
        """This is called by the celery worker. This method calls other helper
        methods to prepare and validate the in and output of a task.
        :param session_data_tasks: List of data task ids from session to check
        access.
        :param args: The dict of arguments submitted to the task.
        :param decrypt: Indicates whether cache must be decrypted to be used.
        :return: The result of the task.
        """
        arguments = self.prepare_args(session_data_tasks, args, decrypt)
        result = self.main(**arguments)
        json = self.task_result_to_json(result)
        return json



"""Provides CategoricalETL for wine quality demo_wine_quality data set."""

import os

import pandas as pd

from fractalis.data.etl import ETL


class CategoricalETL(ETL):
    """CategoricalETL implements support for the
    categorical data of the wine quality data set. """

    name = 'demo_wine_quality_categorical-etl'
    produces = 'categorical'

    @staticmethod
    def can_handle(handler: str, descriptor: dict):
        return handler == 'demo_wine_quality' \
               and descriptor['dataType'] == 'categorical'

    def extract(self, server: str, token: str, descriptor: dict):
        path = os.path.dirname(os.path.abspath(__file__)) + '/wine_quality.csv'
        csv = pd.read_csv(path, sep='\t')
        raw_data = csv[descriptor['field']]
        return raw_data

    def transform(self, raw_data: pd.DataFrame, descriptor: dict):
        df = raw_data.reset_index(level=0)
        df['feature'] = descriptor['field']
        df.columns = ['id', 'value', 'feature']
        df[['id']] = df[['id']].astype(str)
        df[['value']] = df[['value']].astype(str)
        df[['feature']] = df[['feature']].astype(str)
        return df



"""Module containing the Celery task for Boxplot statistics."""

from typing import List, TypeVar
from functools import reduce

import pandas as pd
import numpy as np
import scipy.stats

from fractalis.analytics.task import AnalyticTask
from fractalis.analytics.tasks.shared import utils


T = TypeVar('T')


class BoxplotTask(AnalyticTask):
    """Boxplot Analysis Task implementing AnalyticsTask. This class is a
    submittable celery task."""

    name = 'compute-boxplot'

    def main(self,
             features: List[pd.DataFrame],
             categories: List[pd.DataFrame],
             id_filter: List[T],
             transformation: str,
             subsets: List[List[T]]) -> dict:
        """ Compute boxplot statistics for the given parameters.
        :param features: List of numerical features
        :param categories: List of categorical features used to group numerical
        features.
        :param id_filter: List of ids that will be considered for analysis. If
        empty all ids will be used.
        :param transformation: Transformation that will be applied to the data.
        :param subsets: List of subsets used as another way to group the
        numerical features.
        """
        if not len(features):
            raise ValueError("Must at least specify one "
                             "non empty numerical feature.")
        # merge dfs into single one
        df = reduce(lambda l, r: l.append(r), features)
        df = utils.apply_transformation(df=df, transformation=transformation)
        df.dropna(inplace=True)
        if id_filter:
            df = df[df['id'].isin(id_filter)]
        df = utils.apply_subsets(df=df, subsets=subsets)
        df = utils.apply_categories(df=df, categories=categories)
        df['outlier'] = None
        results = {
            'statistics': {},
            'features': df['feature'].unique().tolist(),
            'categories': df['category'].unique().tolist(),
            'subsets': df['subset'].unique().tolist()
        }
        group_values = []
        for feature in results['features']:
            for subset in results['subsets']:
                for category in results['categories']:
                    values = df[(df['subset'] == subset) &
                                (df['category'] == category) &
                                (df['feature'] == feature)]['value'].tolist()
                    if len(values) < 2:
                        continue
                    # FIXME: v This is ugly. Look at kaplan_meier_survival.py
                    label = '{}//{}//s{}'.format(feature, category, subset + 1)
                    group_values.append(values)
                    stats = self.boxplot_statistics(values)
                    u_outliers = np.array(values) > stats['u_wsk']
                    l_outliers = np.array(values) < stats['l_wsk']
                    outliers = np.bitwise_or(u_outliers, l_outliers)
                    df.loc[(df['subset'] == subset) &
                           (df['category'] == category) &
                           (df['feature'] == feature), 'outlier'] = outliers
                    kde = scipy.stats.gaussian_kde(values)
                    xs = np.linspace(start=stats['l_wsk'],
                                     stop=stats['u_wsk'], num=100)
                    stats['kde'] = kde(xs).tolist()
                    results['statistics'][label] = stats
        results['data'] = df.to_json(orient='records')
        f_value, p_value = scipy.stats.f_oneway(*group_values)
        results['anova'] = {
            'p_value': p_value,
            'f_value': f_value
        }
        return results

    @staticmethod
    def boxplot_statistics(values: List[float]) -> dict:
        """Compute boxplot statistics for the given values.
        :param values: A one dimensional list of numbers.
        :return: A dictionary containing all important boxplot statistics.
        """
        l_qrt = np.percentile(values, 25)
        median = np.percentile(values, 50)
        u_qrt = np.percentile(values, 75)
        iqr = u_qrt - l_qrt
        values = sorted(values)
        # lower whisker as defined by John W. Tukey
        l_wsk = next(value for value in values if value >= l_qrt - 1.5 * iqr)
        values = values[::-1]
        # upper whisker as defined by John W. Tukey
        u_wsk = next(value for value in values if value <= u_qrt + 1.5 * iqr)
        return {
            'l_qrt': l_qrt,
            'median': median,
            'u_qrt': u_qrt,
            'l_wsk': l_wsk,
            'u_wsk': u_wsk
        }



create_data_schema = {
    "type": "object",
    "properties": {
        "handler": {"type": "string"},
        "server": {"type": "string"},
        "auth": {
            "type": "object",
            "properties": {
                "token": {"type": "string"},
                "user": {"type": "string"},
                "passwd": {"type": "string"}
            },
            "minProperties": 1
        },
        "descriptors": {
            "type": "array",
            "items": {
                "type": "object",
                "minProperties": 1
            }
        }
    },
    "required": ["handler", "server", "auth", "descriptors"]
}



from fractalis.utils import list_classes_with_base_class
from fractalis.analytics.task import AnalyticTask

TASK_REGISTRY = list_classes_with_base_class('fractalis.analytics.tasks',
                                             AnalyticTask)



"""The /state controller."""

import re
import json
import logging
import ast
from uuid import UUID, uuid4
from typing import Tuple

from flask import Blueprint, jsonify, Response, request, session

from fractalis import redis, celery
from fractalis.validator import validate_json, validate_schema
from fractalis.analytics.task import AnalyticTask
from fractalis.data.etlhandler import ETLHandler
from fractalis.state.schema import request_state_access_schema, \
    save_state_schema


state_blueprint = Blueprint('state_blueprint', __name__)
logger = logging.getLogger(__name__)


@state_blueprint.route('', methods=['POST'])
@validate_json
@validate_schema(save_state_schema)
def save_state() -> Tuple[Response, int]:
    """Save given payload to redis, so it can be accessed later on.
    :return: UUID linked to the saved state.
    """
    logger.debug("Received POST request on /state.")
    payload = request.get_json(force=True)
    state = str(payload['state'])
    matches = re.findall('\$.+?\$', state)
    task_ids = [AnalyticTask.parse_value(match)[0] for match in matches]
    task_ids = [task_id for task_id in set(task_ids) if task_id is not None]
    if not task_ids:
        error = "This state cannot be saved because it contains no data " \
                "task ids. These are used to verify access to the state and " \
                "its potentially sensitive data."
        logger.error(error)
        return jsonify({'error': error}), 400
    descriptors = []
    for task_id in task_ids:
        value = redis.get('data:{}'.format(task_id))
        if value is None:
            error = "Data task id is {} could not be found in redis. " \
                    "State cannot be saved".format(task_id)
            logger.error(error)
            return jsonify({'error': error}), 400
        data_state = json.loads(value)
        descriptors.append(data_state['meta']['descriptor'])
    assert len(task_ids) == len(descriptors)
    meta_state = {
        'state': ast.literal_eval(state),
        'server': payload['server'],
        'handler': payload['handler'],
        'task_ids': task_ids,
        'descriptors': descriptors
    }
    uuid = uuid4()
    redis.set(name='state:{}'.format(uuid), value=json.dumps(meta_state))
    logger.debug("Successfully saved data to redis. Sending response.")
    return jsonify({'state_id': uuid}), 201


@state_blueprint.route('/<uuid:state_id>', methods=['POST'])
@validate_json
@validate_schema(request_state_access_schema)
def request_state_access(state_id: UUID) -> Tuple[Response, int]:
    """Traverse through the state object linked to the given UUID and look for
    data ids. Then attempt to load the data into the current session to verify
    access.
    :param state_id: The id associated with the saved state.
    :return: See redirect target.
    """
    logger.debug("Received POST request on /state/<uuid:state_id>.")
    wait = request.args.get('wait') == '1'
    payload = request.get_json(force=True)
    state_id = str(state_id)
    value = redis.get('state:{}'.format(state_id))
    if not value:
        error = "Could not find state associated with id {}".format(state_id)
        logger.error(error)
        return jsonify({'error': error}), 404
    meta_state = json.loads(value)
    etl_handler = ETLHandler.factory(handler=meta_state['handler'],
                                     server=meta_state['server'],
                                     auth=payload['auth'])
    task_ids = etl_handler.handle(descriptors=meta_state['descriptors'],
                                  data_tasks=session['data_tasks'],
                                  use_existing=True,
                                  wait=wait)

    session['data_tasks'] += task_ids
    session['data_tasks'] = list(set(session['data_tasks']))
    # if all tasks finish successfully we now that session has access to state
    session['state_access'][state_id] = task_ids
    logger.debug("Tasks successfully submitted. Sending response.")
    return jsonify(''), 202


@state_blueprint.route('/<uuid:state_id>', methods=['GET'])
def get_state_data(state_id: UUID) -> Tuple[Response, int]:
    """Check whether every ETL linked to the state_id successfully executed for
    this session. If and only if every ETL successfully completed grant access
    to the state information.
    :param state_id: ID of the state that is requested.
    :return: Previously saved state.
    """
    logger.debug("Received GET request on /state/<uuid:state_id>.")
    state_id = str(state_id)
    value = redis.get('state:{}'.format(state_id))
    if not value or state_id not in session['state_access']:
        error = "Cannot get state. Make sure to submit a POST request " \
                "to this very same URL containing credentials and server " \
                "data to launch access verification. Only after that a GET " \
                "request might or might not return you the saved state."
        logger.error(error)
        return jsonify({'error': error}), 404
    meta_state = json.loads(value)
    state = json.dumps(meta_state['state'])
    for task_id in session['state_access'][state_id]:
        async_result = celery.AsyncResult(task_id)
        if async_result.state == 'SUBMITTED':
            return jsonify({'message': 'ETLs are still running.'}), 202
        elif async_result.state == 'SUCCESS':
            continue
        else:
            error = "One or more ETLs failed or has unknown status. " \
                    "Assuming no access to saved state."
            logger.error(error)
            return jsonify({'error': error}), 403
    # replace task ids in state with the ids of the freshly loaded data
    for i, task_id in enumerate(meta_state['task_ids']):
        state = re.sub(pattern=task_id,
                       repl=session['state_access'][state_id][i],
                       string=state)
    return jsonify({'state': json.loads(state)}), 200



"""This module provides clustering algorithm for array type data."""

import logging
from typing import List, Tuple
from collections import Counter

import pandas as pd
from scipy.cluster import hierarchy as hclust
from scipy.cluster.vq import kmeans2

from fractalis.analytics.task import AnalyticTask


logger = logging.getLogger(__name__)


class ClusteringTask(AnalyticTask):

    name = 'compute-cluster'

    def main(self, df: dict, cluster_algo: str,
             options: dict) -> dict:
        try:
            df = pd.DataFrame.from_dict(df)
        except Exception:
            error = "Failed to parse input data frame."
            logger.error(error)
            raise ValueError(error)
        # fill NAs with col medians so the clustering algorithms will work
        df = df.T.fillna(df.median(axis=1)).T
        if cluster_algo == 'hclust':
            return self.hclust(df, options)
        elif cluster_algo == 'kmeans':
            return self.kmeans(df, options)
        error = "Algorithm not implemented: '{}'".format(cluster_algo)
        logger.error(error)
        raise ValueError(error)

    def hclust(self, df: pd.DataFrame, options: dict) -> dict:
        try:
            method = options['method']
            metric = options['metric']
            n_row_clusters = options['n_row_clusters']
            n_col_clusters = options['n_col_clusters']
        except KeyError:
            error = "'method', 'metric', 'n_row_clusters', " \
                    "and 'n_col_clusters' are mandatory parameters to " \
                    "perform a hierarchical clustering."
            logger.error(error)
            raise ValueError(error)
        row_names, row_clusters = self._hclust(df, method,
                                               metric, n_row_clusters)
        col_names, col_clusters = self._hclust(df.T, method,
                                               metric, n_col_clusters)
        return {
            'row_clusters': list(zip(row_names, row_clusters)),
            'col_clusters': list(zip(col_names, col_clusters))
        }

    def _hclust(self, df: pd.DataFrame, method: str,
                metric: str, n_clusters: int) -> Tuple[List, List]:
        names = list(df.index)
        values = df.values
        z = hclust.linkage(values, method=method, metric=metric)
        cluster = [x[0] for x in hclust.cut_tree(z,
                                                 n_clusters=[n_clusters])]
        cluster_count = Counter(cluster)
        # sort elements by their cluster size
        sorted_cluster = sorted(zip(names, cluster),
                                key=lambda x: (cluster_count[x[1]], x[1]),
                                reverse=True)
        names = [x[0] for x in sorted_cluster]
        cluster = [x[1] for x in sorted_cluster]
        # relabel cluster, with the biggest cluster being 0
        c = 0
        relabeled_cluster = []
        for i, v in enumerate(cluster):
            if i > 0 and cluster[i] != cluster[i-1]:
                c += 1
            relabeled_cluster.append(c)
        cluster = relabeled_cluster

        return names, cluster

    def kmeans(self, df: pd.DataFrame, options: dict) -> dict:
        try:
            n_row_centroids = options['n_row_centroids']
            n_col_centroids = options['n_col_centroids']
        except KeyError:
            error = "'n_row_centroids' and 'n_col_centroids' are mandatory " \
                    "parameters to perform a kmeans clustering."
            logger.error(error)
            raise ValueError(error)

        row_names, row_clusters = self._kmeans(df, n_row_centroids)
        col_names, col_clusters = self._kmeans(df.T, n_col_centroids)
        return {
            'row_clusters': list(zip(row_names, row_clusters)),
            'col_clusters': list(zip(col_names, col_clusters))
        }

    def _kmeans(self, df: pd.DataFrame, n_centroids) -> Tuple[List, List]:
        names = list(df.index)
        values = df.as_matrix().astype('float')
        cluster = list(kmeans2(values, k=n_centroids, minit='points')[1])
        cluster_count = Counter(cluster)
        # sort elements by their cluster size
        sorted_cluster = sorted(zip(names, cluster),
                                key=lambda x: (cluster_count[x[1]], x[1]),
                                reverse=True)
        names = [x[0] for x in sorted_cluster]
        cluster = [x[1] for x in sorted_cluster]
        # relabel cluster, with the biggest cluster being 0
        c = 0
        relabeled_cluster = []
        for i, v in enumerate(cluster):
            if i > 0 and cluster[i] != cluster[i-1]:
                c += 1
            relabeled_cluster.append(c)
        cluster = relabeled_cluster
        return names, cluster



import time

import pandas as pd
import numpy as np

from fractalis.data.etl import ETL


class RandomDFETL(ETL):

    name = 'test_randomdf_task'
    produces = 'mock'

    @staticmethod
    def can_handle(handler, descriptor):
        return handler == 'test' and \
               descriptor['data_type'] and \
               descriptor['data_type'] == 'default'

    def extract(self, server, token, descriptor):
        if 'fail' in token:
            raise Exception('Throwing because I was told to.')
        fake_raw_data = np.random.randn(10, 5)
        time.sleep(0.5)
        return fake_raw_data

    def transform(self, raw_data, descriptor):
        fake_df = pd.DataFrame(raw_data)
        return fake_df



"""This module provides tests for the heatmap analysis main module."""

import pytest
import pandas as pd
import numpy as np

from fractalis.analytics.tasks.heatmap.main import HeatmapTask


# noinspection PyMissingTypeHints
class TestHeatmap:

    task = HeatmapTask()

    def test_functional(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 6], [102, 'foo', 10],
                          [102, 'bar', 11], [103, 'foo', 15], [103, 'bar', 16],
                          [104, 'foo', 20], [104, 'bar', 21]],
                         columns=['id', 'feature', 'value'])
        ]
        subsets = [[101, 102], [103, 104]]
        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='B',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        assert 'data' in result
        assert 'stats' in result

    def test_functional_with_nans_and_missing(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 5],
                          [102, 'foo', 10],
                          [103, 'foo', float('nan')], [103, 'bar', 15],
                          [104, 'foo', 20], [104, 'bar', 20]],
                         columns=['id', 'feature', 'value'])
        ]
        subsets = [[101, 102], [103, 104]]
        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='B',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        for stat in result['stats']:
            if stat != 'feature' and stat != 'AveExpr':
                assert result['stats'][stat][0] == result['stats'][stat][1]

    def test_main_raises_if_invalid_data(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 6], [102, 'foo', 10],
                          [102, 'bar', 11], [103, 'foo', 15], [103, 'bar', 16],
                          [104, 'foo', 20], [104, 'bar', 21]],
                         columns=['id', 'feature', 'value'])
        ]
        subsets = [[1, 2, 3, 4]]  # does not match sample colnames
        with pytest.raises(ValueError) as e:
            self.task.main(numerical_arrays=numerical_arrays,
                           numericals=[],
                           categoricals=[],
                           ranking_method='mean',
                           params={},
                           id_filter=[],
                           max_rows=100,
                           subsets=subsets)
            assert 'data set is too small' in e

    def test_empty_subset_equals_full_subset(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 6], [102, 'foo', 10],
                          [102, 'bar', 11], [103, 'foo', 15], [103, 'bar', 16],
                          [104, 'foo', 20], [104, 'bar', 21]],
                         columns=['id', 'feature', 'value'])
        ]
        result_1 = self.task.main(numerical_arrays=numerical_arrays,
                                  numericals=[],
                                  categoricals=[],
                                  ranking_method='mean',
                                  params={},
                                  id_filter=[],
                                  max_rows=100,
                                  subsets=[])

        result_2 = self.task.main(numerical_arrays=numerical_arrays,
                                  numericals=[],
                                  categoricals=[],
                                  ranking_method='mean',
                                  params={},
                                  id_filter=[],
                                  max_rows=100,
                                  subsets=[[101, 102, 103, 104]])
        assert result_1 == result_2

    def test_multiple_numerical_array_data(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 6],
                          [102, 'foo', 10], [102, 'bar', 11],
                          [103, 'foo', 15], [103, 'bar', 16],
                          [104, 'foo', 20], [104, 'bar', 21]],
                         columns=['id', 'feature', 'value']),
            pd.DataFrame([[101, 'baz', 10], [102, 'baz', 11],
                          [105, 'foo', 20], [105, 'baz', 21],
                          [106, 'bar', 15]],
                         columns=['id', 'feature', 'value'])
        ]
        subsets = [[101, 102, 106], [103, 104, 105]]
        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='B',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        assert 'data' in result
        assert 'stats' in result

    def test_zscore_is_not_nan_if_data_misses_values(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 6],
                          [102, 'foo', 10], [102, 'bar', 11],
                          [103, 'foo', 15], [103, 'bar', 16],
                          [104, 'foo', 20], [104, 'bar', 21]],
                         columns=['id', 'feature', 'value']),
            pd.DataFrame([[101, 'baz', 10], [102, 'baz', 11],
                          [105, 'foo', 20], [105, 'baz', 21],
                          [106, 'bar', 15]],
                         columns=['id', 'feature', 'value'])
        ]
        subsets = [[101, 102, 106], [103, 104, 105]]
        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='B',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        data = result['data']
        data = pd.DataFrame(data)
        assert not np.isnan(np.min(data['zscore']))

    def test_results_are_sorted(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'A', 5], [102, 'A', 5],
                          [101, 'B', 2], [102, 'B', 2],
                          [101, 'C', 8], [102, 'C', 8],
                          [101, 'D', 10], [102, 'D', 10]],
                         columns=['id', 'feature', 'value'])
        ]
        subsets = []
        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='mean',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        data = result['data']
        data = pd.DataFrame(data)
        feature_col = data['feature'].tolist()
        assert ['D', 'C', 'A', 'B', 'D', 'C', 'A', 'B'] == feature_col
        assert ['D', 'C', 'A', 'B'] == result['stats']['feature']

    def test_max_rows_works(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'A', 5], [102, 'A', 5],
                          [101, 'B', 2], [102, 'B', 2],
                          [101, 'C', 8], [102, 'C', 8],
                          [101, 'D', 10], [102, 'D', 10]],
                         columns=['id', 'feature', 'value'])
        ]
        subsets = []
        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='mean',
                                params={},
                                id_filter=[],
                                max_rows=2,
                                subsets=subsets)
        data = result['data']
        data = pd.DataFrame(data)
        feature_col = data['feature'].tolist()
        assert ['D', 'C', 'D', 'C'] == feature_col
        assert result['stats']['feature'] == ['D', 'C']

    def test_sorts_correct_for_different_criteria(self):
        numerical_arrays = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', -12],
                          [102, 'foo', 10], [102, 'bar', -25],
                          [103, 'foo', 15], [103, 'bar', -20],
                          [104, 'foo', 20], [104, 'bar', -50]],
                         columns=['id', 'feature', 'value'])
        ]
        subsets = [[101, 102], [103, 104]]
        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='P.Value',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        stats = result['stats']['P.Value']
        assert all([stats[i] < stats[i + 1] for i in range(len(stats) - 1)])

        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='adj.P.Val',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        stats = result['stats']['adj.P.Val']
        assert all([stats[i] < stats[i + 1] for i in range(len(stats) - 1)])

        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='B',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        stats = result['stats']['B']
        assert all([stats[i] > stats[i + 1] for i in range(len(stats) - 1)])

        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='logFC',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        stats = result['stats']['logFC']
        assert all([abs(stats[i]) > abs(stats[i + 1])
                    for i in range(len(stats) - 1)])

        result = self.task.main(numerical_arrays=numerical_arrays,
                                numericals=[],
                                categoricals=[],
                                ranking_method='t',
                                params={},
                                id_filter=[],
                                max_rows=100,
                                subsets=subsets)
        stats = result['stats']['t']
        assert all([stats[i] > stats[i + 1] for i in range(len(stats) - 1)])



"""This module provides tests for the misc controller."""

import re

import flask
import pytest


class TestMisc:

    @pytest.fixture(scope='function')
    def test_client(self):
        from fractalis import app
        app.testing = True
        with app.test_client() as test_client:
            yield test_client

    def test_get_version_returns_version(self, test_client):
        rv = test_client.get('/misc/version')
        body = flask.json.loads(rv.get_data())
        assert re.match('^\d+.\d+.\d+$', body['version'])



from time import sleep
from functools import reduce

from fractalis.analytics.task import AnalyticTask


class AddTask(AnalyticTask):

    name = 'add_test_task'

    def main(self, a, b):
        result = {'sum': a + b}
        return result


class DoNothingTask(AnalyticTask):

    name = 'nothing_test_task'

    def main(self, seconds):
        result = {}
        sleep(seconds)
        result['foo'] = 'bar'
        return result


class DivTask(AnalyticTask):

    name = 'div_test_task'

    def main(self, a, b):
        result = {'div': a / b}
        return result


class SumDataFrameTask(AnalyticTask):

    name = 'sum_df_test_task'

    def main(self, a):
        result = {'sum': a.sum().sum()}
        return result


class MergeDataFramesTask(AnalyticTask):

    name = 'merge_df_task'

    def main(self, df_list):
        if not df_list:
            return {'df': ''}
        df = reduce(lambda l, r: l.append(r), df_list)
        return {'df': df.to_json(orient='records')}


class InvalidReturnTask(AnalyticTask):

    name = 'no_dict_task'

    def main(self):
        return 123


class InvalidJSONTask(AnalyticTask):
    name = 'invalid_json_task'

    def main(self):
        return {'a': lambda: 1}



"""This module tests nothing. It's just here for unit test mocking purposes."""

from fractalis.data.check import IntegrityCheck


class MockIntegrityCheck(IntegrityCheck):

    data_type = 'mock'

    def check(self, data: object) -> None:
        pass



"""This module provides tests for the array_stats module."""

import pytest
import pandas as pd


from fractalis.analytics.tasks.shared import array_stats


# noinspection PyMissingOrEmptyDocstring,PyMethodMayBeStatic,PyMissingTypeHints
class TestArrayStats:

    def test_get_limma_stats_raises_for_invalid_subsets(self):
        df = pd.DataFrame([[5, 10, 15, 20]], index=['foo'],
                          columns=[0, 1, 2, 3])
        subsets = [[0, 1]]
        with pytest.raises(ValueError) as e:
            array_stats.get_limma_stats(df=df, subsets=subsets)
            assert 'requires at least two' in e

    def test_get_limma_stats_raises_for_invalid_df(self):
        df = pd.DataFrame([], index=['foo'], columns=[])
        subsets = [[0], [0]]
        with pytest.raises(ValueError) as e:
            array_stats.get_limma_stats(df=df, subsets=subsets)
            assert 'dimension 1x2 or more' in e

    def test_get_limma_stats_returns_correct_for_2_groups(self):
        df = pd.DataFrame([[5, 10, 15, 20]], index=['foo'],
                          columns=[0, 1, 2, 3])
        subsets = [[0, 1], [2, 3]]
        stats = array_stats.get_limma_stats(df=df, subsets=subsets)
        assert all(stat in list(stats) for stat in
                   ['feature', 'logFC', 'AveExpr', 't', 'P.Value', 'adj.P.Val',
                    'B'])

    def test_get_limma_stats_returns_correct_for_3_groups(self):
        df = pd.DataFrame([[5, 10, 15, 20]], index=['foo'],
                          columns=[0, 1, 2, 3])
        subsets = [[0, 1], [2], [3]]
        stats = array_stats.get_limma_stats(df=df, subsets=subsets)
        assert all(stat in list(stats) for stat in
                   ['feature', 'AveExpr', 'F', 'P.Value', 'adj.P.Val'])
        assert all(stat not in list(stats) for stat in ['logFC', 'B', 't'])

    def test_get_limma_stats_returns_correct_for_4_groups(self):
        df = pd.DataFrame([[5, 10, 15, 20]], index=['foo'],
                          columns=[0, 1, 2, 3])
        subsets = [[0, 1], [1, 2], [2, 3], [3, 0]]
        stats = array_stats.get_limma_stats(df=df, subsets=subsets)
        assert all(stat in list(stats) for stat in
                   ['feature', 'AveExpr', 'F', 'P.Value', 'adj.P.Val'])
        assert all(stat not in list(stats) for stat in ['logFC', 'B', 't'])

    def test_get_deseq2_stats_returns_correct_for_2_groups(self):
        df = pd.DataFrame([[500, 1, 1, 500],
                           [1, 500, 500, 1]],
                          columns=[0, 1, 2, 3])
        subsets = [[0, 1], [2, 3]]
        stats = array_stats.get_deseq2_stats(df=df, subsets=subsets)
        assert all(stat in list(stats) for stat in
                   ['baseMean', 'log2FoldChange', 'lfcSE',
                    'stat', 'pvalue', 'padj'])

    def test_deseq2_requires_exactly_2_subsets(self):
        with pytest.raises(ValueError):
            array_stats.get_deseq2_stats(df=pd.DataFrame(),
                                         subsets=[[], [], []])
        with pytest.raises(ValueError):
            array_stats.get_deseq2_stats(df=pd.DataFrame(), subsets=[[]])



save_state_schema = {
    "type": "object",
    "properties": {
        "state": {"type": "object"},
        "handler": {"type": "string"},
        "server": {"type": "string"}
    },
    "required": ["handler", "server", "state"]
}


request_state_access_schema = {
    "type": "object",
    "properties": {
        "auth": {
            "type": "object",
            "properties": {
                "token": {"type": "string"},
                "user": {"type": "string"},
                "passwd": {"type": "string"}
            },
            "minProperties": 1
        }
    },
    "required": ["auth"]
}



import os

from fractalis import app, redis, sync, celery


@celery.task
def janitor():
    """Ideally this is maintained by a systemd service to cleanup redis and the
    file system while Fractalis is running.
    """
    data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
    if not os.path.exists(data_dir):
        return
    cached_files = [f for f in os.listdir(data_dir)
                    if os.path.isfile(os.path.join(data_dir, f))]
    tracked_files = [key.split(':')[1] for key in redis.scan_iter('data:*')]

    # clean cached files
    for cached_file in cached_files:
        if cached_file not in tracked_files:
            sync.remove_file(os.path.join(data_dir, cached_file))

    # clean tracked files
    for task_id in tracked_files:
        path = os.path.join(data_dir, task_id)
        async_result = celery.AsyncResult(task_id)
        if async_result.state == 'SUCCESS' and not os.path.exists(path):
            redis.delete('data:{}'.format(task_id))



"""This module provides tests for the cluster task
within the heatmap workflow."""

import pytest

from fractalis.analytics.tasks.heatmap.cluster import ClusteringTask


# noinspection PyMissingOrEmptyDocstring,PyMethodMayBeStatic
class TestClustering:

    task = ClusteringTask()

    df = {
        'A': {
            'a': 50,
            'b': 2,
            'c': 45
        },
        'B': {
            'a': 250,
            'b': 5,
            'c': 300
        },
        'C': {
            'a': 55,
            'b': 4,
            'c': 60
        }
    }

    def test_hclust_raises_with_invalid_param_1(self):
        with pytest.raises(ValueError) as e:
            options = {
                'method': 'abc',
                'metric': 'euclidean',
                'n_row_clusters': 2,
                'n_col_clusters': 2
            }
            self.task.main(df=self.df, cluster_algo='hclust', options=options)
            assert 'Invalid method' in e

    def test_hclust_raises_with_invalid_param_2(self):
        with pytest.raises(ValueError) as e:
            options = {
                'method': 'single',
                'metric': 'abc',
                'n_row_clusters': 2,
                'n_col_clusters': 2
            }
            self.task.main(df=self.df, cluster_algo='hclust', options=options)
            assert 'Invalid metric' in e

    def test_hclust_raises_with_invalid_param_3(self):
        with pytest.raises(ValueError) as e:
            options = {
                'method': 'single',
                'metric': 'abc',
                'n_row_clusters': 2,
            }
            self.task.main(df=self.df, cluster_algo='hclust', options=options)
            assert 'mandatory parameters' in e

    def test_hclust_can_handle_identical_cluster_size(self):
        df = {
            'A': {
                'a': 5, 'b': 10
            },
            'B': {
                'a': 500, 'b': 550
            },
            'C': {
                'a': 5, 'b': 10
            },
            'D': {
                'a': 500, 'b': 550
            }
        }
        options = {
            'method': 'single',
            'metric': 'euclidean',
            'n_row_clusters': 2,
            'n_col_clusters': 2
        }
        result = self.task.main(df=df, cluster_algo='hclust', options=options)
        assert ['B', 'D', 'A', 'C'] == [x[0] for x in result['col_clusters']]
        assert [0, 0, 1, 1] == [x[1] for x in result['col_clusters']]

    def test_hclust_returns_valid_result(self):
        options = {
            'method': 'single',
            'metric': 'euclidean',
            'n_row_clusters': 2,
            'n_col_clusters': 2
        }
        result = self.task.main(df=self.df,
                                cluster_algo='hclust', options=options)
        assert 'row_clusters' in result
        assert 'col_clusters' in result
        assert ['a', 'c', 'b'] == [x[0] for x in result['row_clusters']]
        assert ['A', 'C', 'B'] == [x[0] for x in result['col_clusters']]
        assert [0, 0, 1] == [x[1] for x in result['col_clusters']]
        assert [0, 0, 1] == [x[1] for x in result['col_clusters']]

    def test_kmean_raises_with_invalid_param_1(self):
        with pytest.raises(ValueError) as e:
            options = {
                'n_row_centroids': 2,
                'n_col_centroids': 'abc'
            }
            self.task.main(df=self.df, cluster_algo='kmeans', options=options)
            assert 'invalid' in e

    def test_kmean_raises_with_invalid_param_2(self):
        with pytest.raises(ValueError) as e:
            options = {
                'n_row_centroids': 2,
            }
            self.task.main(df=self.df, cluster_algo='kmeans', options=options)
            assert 'mandatory parameters' in e

    def test_kmeans_can_handle_identical_cluster_size(self):
        df = {
            'A': {
                'a': 5, 'b': 10
            },
            'B': {
                'a': 500, 'b': 550
            },
            'C': {
                'a': 5, 'b': 10
            },
            'D': {
                'a': 500, 'b': 550
            }
        }
        options = {
            'n_row_centroids': 2,
            'n_col_centroids': 2
        }
        result = self.task.main(df=df, cluster_algo='kmeans', options=options)
        assert [0, 0, 1, 1] == [x[1] for x in result['col_clusters']]

    def test_kmean_returns_valid_result(self):
        options = {
            'n_row_centroids': 2,
            'n_col_centroids': 2
        }
        result = self.task.main(df=self.df,
                                cluster_algo='kmeans', options=options)
        assert 'row_clusters' in result
        assert 'col_clusters' in result
        assert ['a', 'c', 'b'] == [x[0] for x in result['row_clusters']]
        assert ['A', 'C', 'B'] == [x[0] for x in result['col_clusters']]
        assert [0, 0, 1] == [x[1] for x in result['col_clusters']]
        assert [0, 0, 1] == [x[1] for x in result['col_clusters']]



"""Provides numerical concept ETL for tranSMART."""

from fractalis.data.etls.transmart.shared import create_etl_type, NUMERICAL_FIELD

NumericalETL = create_etl_type(
    name_='transmart_numerical_etl',
    produces_='numerical',
    field_name=NUMERICAL_FIELD
)



"""Provides CategoricalETL for PIC-SURE API."""

import json
from typing import List

import pandas as pd
from io import StringIO

from fractalis.data.etl import ETL
from fractalis.data.etls.picsure import shared


class CategoricalETL(ETL):
    """CategoricalETL implements support for PIC-SURE 'categorical' type."""

    name = 'pic-sure_categorical_etl'
    produces = 'categorical'

    @staticmethod
    def can_handle(handler: str, descriptor: dict) -> bool:
        return handler == 'pic-sure' and \
               descriptor['dataType'] == 'categorical'

    def extract(self, server: str, token: str, descriptor: dict) -> List[str]:
        result_id = shared.submit_query(
            query=json.dumps(descriptor['query']), server=server, token=token)
        shared.wait_for_completion(
            result_id=result_id, server=server, token=token)
        raw_data = shared.get_data(
            result_id=result_id, server=server, token=token)
        return raw_data

    def transform(self, raw_data: List[dict],
                  descriptor: dict) -> pd.DataFrame:
        df = pd.read_csv(StringIO(raw_data))
        feature = df.columns[1]
        df.columns = ['id', 'value']
        df.insert(1, 'feature', feature)
        df[['id']] = df[['id']].astype(str)
        df[['value']] = df[['value']].astype(str)
        return df



"""Test suite for correlation analysis."""
import json

import pytest
import pandas as pd
import numpy as np

from fractalis.analytics.tasks.correlation.main import CorrelationTask


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestCorrelation:

    task = CorrelationTask()

    def test_correct_output(self):
        x = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3],
                          [104, 'foo', 4]], columns=['id', 'feature', 'value'])
        y = pd.DataFrame([[101, 'bar', 1], [102, 'bar', 2], [103, 'bar', 3],
                          [104, 'bar', 4]], columns=['id', 'feature', 'value'])
        result = self.task.main(x=x,
                                y=y,
                                id_filter=[],
                                method='pearson',
                                subsets=[],
                                categories=[])
        assert result['coef'] == 1
        assert result['p_value'] == 0
        assert result['slope']
        assert result['intercept']
        assert result['method'] == 'pearson'
        assert result['data']
        assert result['x_label'] == 'foo'
        assert result['y_label'] == 'bar'
        data = json.loads(result['data'])
        assert 'id' in data[0]
        assert 'subset' in data[0]
        assert 'feature_x' in data[0]
        assert 'feature_y' in data[0]
        assert 'value_x' in data[0]
        assert 'value_y' in data[0]
        assert 'category' in data[0]

    def test_correct_shape(self):
        x = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3],
                          [104, 'foo', 4]], columns=['id', 'feature', 'value'])
        y = pd.DataFrame([[101, 'bar', 1], [102, 'bar', 2], [103, 'bar', 3],
                          [104, 'bar', 4]], columns=['id', 'feature', 'value'])
        result = self.task.main(x=x,
                                y=y,
                                id_filter=[101, 102, 104],
                                method='pearson',
                                subsets=[[101, 102, 103], [102, 103, 104]],
                                categories=[])
        df = pd.DataFrame(json.loads(result['data']))
        assert df.shape == (4, 7)

    def test_empty_subset_equals_full_subset(self):
        x = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3],
                          [104, 'foo', 4]], columns=['id', 'feature', 'value'])
        y = pd.DataFrame([[101, 'bar', 1], [102, 'bar', 2], [103, 'bar', 3],
                          [104, 'bar', 4]], columns=['id', 'feature', 'value'])
        result_1 = self.task.main(x=x,
                                  y=y,
                                  id_filter=[101, 102, 104],
                                  method='pearson',
                                  subsets=[[101, 102, 103, 104]],
                                  categories=[])
        result_2 = self.task.main(x=x,
                                  y=y,
                                  id_filter=[101, 104, 102],
                                  method='pearson',
                                  subsets=[],
                                  categories=[])
        assert result_1 == result_2

    def test_raises_for_unknown_method(self):
        x = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3],
                          [104, 'foo', 4]], columns=['id', 'feature', 'value'])
        y = pd.DataFrame([[101, 'bar', 1], [102, 'bar', 2], [103, 'bar', 3],
                          [104, 'bar', 4]], columns=['id', 'feature', 'value'])
        with pytest.raises(ValueError) as e:
            self.task.main(x=x,
                           y=y,
                           id_filter=[],
                           method='foo',
                           subsets=[],
                           categories=[])
            assert 'Unknown method' in e

    def test_x_and_y_intersect_must_be_non_empty(self):
        x = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3],
                          [104, 'foo', 4]], columns=['id', 'feature', 'value'])
        y = pd.DataFrame([[201, 'bar', 1], [202, 'bar', 2], [203, 'bar', 3],
                          [204, 'bar', 4]], columns=['id', 'feature', 'value'])
        with pytest.raises(ValueError) as e:
            self.task.main(x=x,
                           y=y,
                           id_filter=[],
                           method='pearson',
                           subsets=[],
                           categories=[])
            assert 'do not share any ids' in e

    def test_returns_nans_if_too_few_values(self):
        x = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3],
                          [104, 'foo', 4]], columns=['id', 'feature', 'value'])
        y = pd.DataFrame([[101, 'bar', 1], [102, 'bar', 2], [103, 'bar', 3],
                          [104, 'bar', 4]], columns=['id', 'feature', 'value'])
        result = self.task.main(x=x,
                                y=y,
                                id_filter=[],
                                method='pearson',
                                subsets=[[101], [102, 104], [103], []],
                                categories=[])
        assert not np.isnan(result['coef'])



import os
import inspect
import importlib
from pathlib import Path
from typing import List
import importlib.util


def import_module_by_abs_path(module_path: str) -> object:
    """Import the module for the given path.
    :param module_path: The absolute path to the module.
    :return: A reference to the imported module.
    """
    module_name = os.path.splitext(os.path.basename(module_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def list_classes_with_base_class(
        package: str, base_class: object) -> List[object]:
    """For the given base_class list classes in the given package that do
    implement it.
    :param package: The package to search.
    :param base_class: The base class.
    :return: A list of classes that implement the base class.
    """
    package = importlib.import_module(package)
    abs_path = os.path.dirname(os.path.abspath(package.__file__))
    class_list = []
    for f in Path(abs_path).glob('*/**/*.py'):
        module_path = str(f)
        if not os.path.basename(module_path).startswith('_'):
            module = import_module_by_abs_path(module_path)
            classes = inspect.getmembers(module, inspect.isclass)
            for name, obj in classes:
                if obj.__base__ == base_class:
                    class_list.append(obj)
    return class_list


def get_cache_encrypt_key(key):
    """Prepare key for use with crypto libs.
    :param key: Passphrase used for encryption.
    """
    key += (16 - (len(key) % 16)) * '-'
    return key.encode('utf-8')



"""This module provides test for the pca task."""

import pandas as pd
import numpy as np

from fractalis.analytics.tasks.pca.main import PCATask


# noinspection PyMissingTypeHints
class TestPCATask:

    task = PCATask()

    def test_correct_output(self):
        features = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 6],
                          [102, 'foo', 10], [102, 'bar', 11],
                          [103, 'foo', 15], [103, 'bar', 16],
                          [104, 'foo', 20], [104, 'bar', 21]],
                         columns=['id', 'feature', 'value']),
            pd.DataFrame([[101, 'baz', 5],
                          [102, 'baz', 10],
                          [104, 'baz', 20],
                          [105, 'baz', 100]],
                         columns=['id', 'feature', 'value'])
        ]
        categories = [
            pd.DataFrame([[101, '_', 'a'],
                          [102, '_', 'a'],
                          [104, '_', 'a']],
                         columns=['id', 'feature', 'value'])
        ]
        result = self.task.main(features=features,
                                categories=categories,
                                whiten=False,
                                id_filter=[],
                                subsets=[])
        data = result['data']
        assert 0 in data
        assert 1 in data
        assert 'category' in data
        assert 'subset' in data
        assert 'id' in data
        assert data['id'] == [101, 102, 103, 104, 105]
        assert data['subset'] == [0, 0, 0, 0, 0]
        np.testing.assert_equal(data['category'],
                                ['a', 'a', float('nan'), 'a', float('nan')])

    def test_id_filter_works(self):
        features = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 6],
                          [102, 'foo', 10], [102, 'bar', 11],
                          [103, 'foo', 15], [103, 'bar', 16],
                          [104, 'foo', 20], [104, 'bar', 21]],
                         columns=['id', 'feature', 'value'])
        ]
        result = self.task.main(features=features,
                                categories=[],
                                whiten=False,
                                id_filter=[101, 104],
                                subsets=[])
        data = result['data']
        assert all(np.unique(data['id']) == [101, 104])

    def test_correct_loadings(self):
        features = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 20],
                          [102, 'foo', 10], [102, 'bar', 15],
                          [103, 'foo', 15], [103, 'bar', 10],
                          [104, 'foo', 20], [104, 'bar', 5]],
                         columns=['id', 'feature', 'value'])
        ]
        result = self.task.main(features=features,
                                categories=[],
                                whiten=False,
                                id_filter=[],
                                subsets=[])
        loadings = result['loadings']
        assert loadings[0][0] == -loadings[0][1]
        assert loadings[1][0] == loadings[1][1]

    def test_correct_variance_ratios(self):
        features = [
            pd.DataFrame([[101, 'foo', 5], [101, 'bar', 5],
                          [102, 'foo', 10], [102, 'bar', 5],
                          [103, 'foo', 15], [103, 'bar', 5],
                          [104, 'foo', 20], [104, 'bar', 5]],
                         columns=['id', 'feature', 'value'])
        ]
        result = self.task.main(features=features,
                                categories=[],
                                whiten=False,
                                id_filter=[],
                                subsets=[])
        variance_ratios = result['variance_ratios']
        assert variance_ratios == [1, 0]



"""This module provides statistics for volcano plots."""

import logging
from typing import List
from functools import reduce

import pandas as pd

from fractalis.analytics.task import AnalyticTask
from fractalis.analytics.tasks.shared import utils, array_stats

# TODO: Log more
logger = logging.getLogger(__name__)


class VolcanoTask(AnalyticTask):
    """Volcanoplot Analysis Task implementing AnalyticsTask. This class is a
    submittable celery task."""

    name = 'compute-volcanoplot'

    def main(self, numerical_arrays: List[pd.DataFrame],
             id_filter: List[str],
             ranking_method: str,
             params: dict,
             subsets: List[List[str]]) -> dict:
        # TODO: docstring
        # merge input data into single df
        df = reduce(lambda a, b: a.append(b), numerical_arrays)
        if not subsets:
            # empty subsets equals all samples in one subset
            subsets = [df['id'].unique().tolist()]
        else:
            # if subsets are defined we drop the rows that are not part of one
            flattened_subsets = [x for subset in subsets for x in subset]
            df = df[df['id'].isin(flattened_subsets)]
        # apply id filter
        if id_filter:
            df = df[df['id'].isin(id_filter)]
        # drop subset ids that are not in the df
        subsets = utils.drop_unused_subset_ids(df=df, subsets=subsets)
        # make sure the input data are still valid after the pre-processing
        if df.shape[0] < 1:
            error = "Either the input data set is too small or " \
                    "the subset sample ids do not match the data."
            logger.error(error)
            raise ValueError(error)
        # make matrix of input data
        df = df.pivot(index='feature', columns='id', values='value')
        # compute the stats (p / fC) for the selected ranking method
        stats = array_stats.get_stats(df=df,
                                      subsets=subsets,
                                      params=params,
                                      ranking_method=ranking_method)
        return {
            'stats': stats.to_dict(orient='list')
        }



"""This module provides the ETL class"""

import abc
import json
import logging
import os

from Cryptodome.Cipher import AES
# noinspection PyProtectedMember
from celery import Task
from pandas import DataFrame

from fractalis import app, redis
from fractalis.data.check import IntegrityCheck
from fractalis.utils import get_cache_encrypt_key

logger = logging.getLogger(__name__)


class ETL(Task, metaclass=abc.ABCMeta):
    """This is an abstract class that implements a celery Task and provides a
    factory method to create instances of implementations of itself. Its main
    purpose is to  manage extraction of the data from the target server. ETL
    stands for (E)xtract (T)ransform (L)oad and not by coincidence similar
    named methods can be found in this class.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Used by celery to identify this task by name."""
        pass

    @property
    @abc.abstractmethod
    def produces(self) -> str:
        """This specifies the fractalis internal format that this ETL
        produces. Can be one of: ['categorical', 'numerical', 'numerical_array']
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def can_handle(handler: str, descriptor: dict) -> bool:
        """Check if the current implementation of ETL can handle given handler
        and data type.
        WARNING: You should never raise an Exception here and expect it to be
        propagated further up. It will be caught and assumed that the
        current ETL cannot handle the given arguments.
        :param handler: Describes the handler. E.g.: transmart, ada
        :param descriptor: Describes the data that we want to download.
        :return: True if implementation can handle given parameters.
        """
        pass

    @staticmethod
    def factory(handler: str, descriptor: dict) -> 'ETL':
        """Return an instance of the implementation ETL that can handle the
        given parameters.
        :param handler: Describes the handler. E.g.: transmart, ada
        :param descriptor: Describes the data that we want to download.
        :return: An instance of an implementation of ETL that returns True for
        can_handle()
        """
        from . import ETL_REGISTRY
        for ETLTask in ETL_REGISTRY:
            # noinspection PyBroadException
            try:
                if ETLTask.can_handle(handler, descriptor):
                    return ETLTask()
            except Exception as e:
                logger.warning("Caught exception and assumed that ETL '{}' "
                               "cannot handle handler '{}' "
                               "and descriptor: '{}'. Exception:'{}'".format(
                                   type(ETLTask).__name__,
                                   handler,
                                   str(descriptor), e))
                continue

        raise NotImplementedError(
            "No ETL implementation found for handler '{}' "
            "and descriptor '{}'".format(handler, descriptor))

    @abc.abstractmethod
    def extract(self, server: str, token: str, descriptor: dict) -> object:
        """Extract the data via HTTP requests.
        :param server: The server from which to extract from.
        :param token: The token used for authentication.
        :param descriptor: The descriptor containing all necessary information
        to download the data.
        """
        pass

    @abc.abstractmethod
    def transform(self, raw_data: object, descriptor: dict) -> DataFrame:
        """Transform the data into a pandas.DataFrame with a naming according to
        the Fractalis standard format.
        :param raw_data: The return value of extract().
        :param descriptor: The data descriptor, sometimes needed
        for transformation
        """
        pass

    def sanity_check(self):
        """Check whether ETL is still sane and should be continued. E.g. if
        redis has been cleared it does not make sense to proceed. Raise an
        exception if not sane."""
        check_1 = redis.exists('data:{}'.format(self.request.id))
        if not check_1:
            error = "ETL failed! The associated entry in " \
                    "Redis has been removed while the ETL was running."
            logger.error(error)
            raise RuntimeError(error)

    def update_redis(self, data_frame: DataFrame) -> None:
        """Set several meta information that can be used to filter the data
        before the analysis.
        :param data_frame: The extracted and transformed data.
        """
        value = redis.get(name='data:{}'.format(self.request.id))
        assert value is not None
        data_state = json.loads(value)
        if 'feature' in data_frame.columns:
            features = data_frame['feature'].unique().tolist()
        else:
            features = []
        data_state['meta']['features'] = features
        redis.setex(name='data:{}'.format(self.request.id),
                    value=json.dumps(data_state),
                    time=app.config['FRACTALIS_DATA_LIFETIME'])

    @staticmethod
    def secure_load(data_frame: DataFrame, file_path: str) -> None:
        """Save data to the file system in encrypted form using AES and the
        web service secret key. This can be useful to comply with certain
        security standards.
        :param data_frame: DataFrame to write.
        :param file_path: File to write to.
        """
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        data = data_frame.to_json()
        data = data.encode('utf-8')
        key = get_cache_encrypt_key(app.config['SECRET_KEY'])
        cipher = AES.new(key, AES.MODE_EAX)
        ciphertext, tag = cipher.encrypt_and_digest(data)
        with open(file_path, 'wb') as f:
            [f.write(x) for x in (cipher.nonce, tag, ciphertext)]

    @staticmethod
    def load(data_frame: DataFrame, file_path: str) -> None:
        """Load (save) the data to the file system.
        :param data_frame: DataFrame to write.
        :param file_path: File to write to.
        """
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        data_frame.to_pickle(file_path, compression='gzip')

    def run(self, server: str, token: str,
            descriptor: dict, file_path: str,
            encrypt: bool) -> None:
        """Run extract, transform and load. This is called by the celery worker.
        This is called by the celery worker.
        :param
        :param server: The server on which the data are located.
        :param token: The token used for authentication.
        :param descriptor: Contains all necessary information to download data
        :param file_path: The location where the data will be stored
        :param encrypt: Whether or not the data should be encrypted.
        :return: The data id. Used to access the associated redis entry later
        """
        logger.info("Starting ETL process ...")
        logger.info("(E)xtracting data from server '{}'.".format(server))
        try:
            self.sanity_check()
            raw_data = self.extract(server, token, descriptor)
        except Exception as e:
            logger.exception(e)
            raise RuntimeError("Data extraction failed. {}".format(e))
        logger.info("(T)ransforming data to Fractalis format.")
        try:
            self.sanity_check()
            data_frame = self.transform(raw_data, descriptor)
            checker = IntegrityCheck.factory(self.produces)
            checker.check(data_frame)
        except Exception as e:
            logger.exception(e)
            raise RuntimeError("Data transformation failed. {}".format(e))
        if not isinstance(data_frame, DataFrame):
            error = "transform() must return 'pandas.DataFrame', " \
                    "but returned '{}' instead.".format(type(data_frame))
            logging.error(error, exc_info=1)
            raise TypeError(error)
        try:
            self.sanity_check()
            if encrypt:
                self.secure_load(data_frame, file_path)
            else:
                self.load(data_frame, file_path)
            self.update_redis(data_frame)
        except Exception as e:
            logger.exception(e)
            raise RuntimeError("Data loading failed. {}".format(e))



"""This module contains common functions used in analytic tasks."""

import logging
from typing import List, TypeVar
from functools import reduce
from copy import deepcopy

import pandas as pd
import numpy as np


logger = logging.getLogger(__name__)
T = TypeVar('T')


def apply_subsets(df: pd.DataFrame,
                  subsets: List[List[T]]) -> pd.DataFrame:
    """Build a new DataFrame that contains a new column 'subset' defining
    the subset the data point belongs to. If a data point belongs to
    multiple subsets then the row is duplicated.
    :param df: The DataFrame used as a base.
    :param subsets: The subsets defined by the user.
    :return: The new DataFrame with an additional 'subset' column.
    """
    if not subsets:
        subsets = [df['id']]
    _df = pd.DataFrame()
    for i, subset in enumerate(subsets):
        df_subset = df[df['id'].isin(subset)]
        if not df_subset.shape[0]:
            continue
        subset_col = [i] * df_subset.shape[0]
        df_subset = df_subset.assign(subset=subset_col)
        _df = _df.append(df_subset)
    if _df.shape[0] == 0:
        raise ValueError("No data match given subsets.")
    return _df


def apply_categories(df: pd.DataFrame,
                     categories: List[pd.DataFrame]) -> pd.DataFrame:
    """Collapse all category DataFrames into a single column and add it as a new
     column to the given DataFrame. This is done by joining all columns values
    via the string `&&`
    :param df: The DataFrame to be extended
    :param categories: List of category DataFrames
    :return: The base DataFrame with an additional 'category' column
    """
    if len(categories):
        # drop 'feature' column from dfs
        categories = [df.drop('feature', axis=1) for df in categories]
        # merge all dfs into one
        data = reduce(lambda l, r: l.merge(r, on='id', how='outer'),
                      categories)
        # remember ids
        ids = data['id']
        # drop id column
        data = data.drop('id', axis=1)
        # replace everything that is not an category with ''
        data = data.applymap(
            lambda el: el if isinstance(el, str) and el else '')
        # join all columns with && into a single one. Ignore '' entries.
        data = data.apply(
            lambda row: ' AND '.join(
                list(map(str, [el for el in row.tolist() if el]))), axis=1)
        # cast Series to DataFrame
        data = pd.DataFrame(data, columns=['category'])
        # reassign ids to collapsed df
        data = data.assign(id=ids)
        # merge category data into main df
        df = df.merge(data, on='id', how='left')
        # get unique categories
    else:
        df = df.assign(category='')
    return df


def drop_unused_subset_ids(df: pd.DataFrame,
                           subsets: List[List[T]]) -> List[List[T]]:
    """Drop subset ids that are not present in the given data
    :param df: Dataframe containing array data in the Fractalis format.
    :param subsets: Subset groups specified by the user.
    :return: Modified subsets list.
    """
    ids = df['id'].unique()
    _subsets = deepcopy(subsets)
    for subset in _subsets:
        _subset = list(subset)
        for id in _subset:
            if id not in ids:
                subset.remove(id)
    return _subsets


def apply_transformation(df: pd.DataFrame, transformation: str) -> pd.DataFrame:
    """Apply transformation to the value column of the data frame.
    E.g. log2 or 10^x scales.
    NaN and Inf are dropped!
    :param df: Dataframe containing array data in the Fractalis format.
    :param transformation: The transformation to apply.
    :return: The dataframe with an transformed value column excl. NaN and Inf
    """
    transformations = {
        'identity': lambda x: x,
        'log2(x)': np.log2,
        'log10(x)': np.log10,
        '2^x': lambda x: np.power(2.0, x),
        '10^x': lambda x: np.power(10.0, x)
    }
    # drop zeros because
    df = df[df['value'] != 0]
    df['value'] = transformations[transformation](df['value'])
    if np.any(np.isinf(df['value'])):
        error = 'Found inf after transformation. Transformation "{}" should ' \
                'only be used on log scaled data.'.format(transformation)
        logger.exception(error)
        raise ValueError(error)
    return df



"""This module contains tests for the common module in the shared package."""

import pytest
import pandas as pd
import numpy as np

from fractalis.analytics.tasks.shared import utils


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestCommonTasks:

    def test_apply_subsets(self):
        df = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3]],
                          columns=['id', 'feature', 'value'])
        subsets = [[101, 102], [], [103, 102, 104]]
        result = utils.apply_subsets(df=df, subsets=subsets)
        assert result['subset'].tolist() == [0, 0, 2, 2]

    def test_apply_categorys(self):
        df = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3]],
                          columns=['id', 'feature', 'value'])
        c1 = pd.DataFrame([[101, 'c1', 'a'],
                           [102, 'c1', 'b'],
                           [105, 'c1', 'c']],
                          columns=['id', 'feature', 'value'])
        c2 = pd.DataFrame([[106, 'c2', 'd']],
                          columns=['id', 'feature', 'value'])
        c3 = pd.DataFrame([[102, 'c3', 'f']],
                          columns=['id', 'feature', 'value'])
        result = utils.apply_categories(df=df, categories=[c1, c2, c3])
        assert result['category'].tolist()[:2] == ['a', 'b AND f']
        assert np.isnan(result['category'].tolist()[2])

    def test_drop_unused_subset_ids(self):
        df = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3]],
                          columns=['id', 'feature', 'value'])
        subsets = [[], [101], [103, 104], [105]]
        subsets = utils.drop_unused_subset_ids(df=df, subsets=subsets)
        assert subsets == [[], [101], [103], []]

    def test_apply_transformation_raises_if_pos_inf(self):
        df = pd.DataFrame([[101, 'foo', 1000]],
                          columns=['id', 'feature', 'value'])
        with pytest.raises(ValueError) as e:
            utils.apply_transformation(df=df, transformation='10^x')
            assert 'only be used on log scaled data' in e

    def test_apply_transformation_drops_zeros(self):
        df = pd.DataFrame([[101, 'foo', 1000], [102, 'foo', 0]],
                          columns=['id', 'feature', 'value'])
        assert df.shape[0] == 2
        df = utils.apply_transformation(df=df, transformation='log2(x)')
        assert df.shape[0] == 1
